=== modified file 'bzrlib/chk_map.py'
--- bzrlib/chk_map.py	2008-10-01 03:46:28 +0000
+++ bzrlib/chk_map.py	2008-10-09 04:50:37 +0000
@@ -81,12 +81,18 @@
             result._map(key, value)
         return result._save()
 
-    def iteritems(self):
+    def iteritems(self, key_filter=None):
         """Iterate over the entire CHKMap's contents."""
         self._ensure_root()
-        for name, key, in self._root_node._nodes.iteritems():
-            bytes = self._read_bytes(key)
-            yield name, ValueNode.deserialise(bytes).value
+        if key_filter is not None:
+            for name, key, in self._root_node._nodes.iteritems():
+                if name in key_filter:
+                    bytes = self._read_bytes(key)
+                    yield name, ValueNode.deserialise(bytes).value
+        else:
+            for name, key, in self._root_node._nodes.iteritems():
+                bytes = self._read_bytes(key)
+                yield name, ValueNode.deserialise(bytes).value
 
     def _map(self, key, value):
         """Map key to value."""
@@ -152,6 +158,10 @@
         self._nodes = nodes
         self._key = key
 
+    def refs(self):
+        """Get the CHK key references this node holds."""
+        return self._nodes.values()
+
     def remove_child(self, name):
         """Remove name from the node.
 

=== added file 'bzrlib/chk_serializer.py'
--- bzrlib/chk_serializer.py	1970-01-01 00:00:00 +0000
+++ bzrlib/chk_serializer.py	2008-10-09 04:50:37 +0000
@@ -0,0 +1,59 @@
+# Copyright (C) 2008 Canonical Ltd
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+
+"""Serializer object for CHK based inventory storage."""
+
+from bzrlib import (
+    inventory,
+    xml5,
+    xml6,
+    )
+
+
+class CHKSerializerSubtree(xml6.Serializer_v6):
+    """A CHKInventory based serializer that supports tree references"""
+
+    supported_kinds = set(['file', 'directory', 'symlink', 'tree-reference'])
+    format_num = '9'
+    revision_format_num = None
+    support_altered_by_hack = False
+
+    def _unpack_entry(self, elt):
+        kind = elt.tag
+        if not kind in self.supported_kinds:
+            raise AssertionError('unsupported entry kind %s' % kind)
+        if kind == 'tree-reference':
+            file_id = elt.attrib['file_id']
+            name = elt.attrib['name']
+            parent_id = elt.attrib['parent_id']
+            revision = elt.get('revision')
+            reference_revision = elt.get('reference_revision')
+            return inventory.TreeReference(file_id, name, parent_id, revision,
+                                           reference_revision)
+        else:
+            return xml6.Serializer_v6._unpack_entry(self, elt)
+
+
+class CHKSerializer(xml5.Serializer_v5):
+    """A CHKInventory based serializer with 'plain' behaviour."""
+
+    format_num = '9'
+    revision_format_num = None
+    support_altered_by_hack = False
+
+
+chk_serializer_subtree = CHKSerializerSubtree()
+chk_serializer = CHKSerializer()

=== modified file 'bzrlib/fetch.py'
--- bzrlib/fetch.py	2008-08-29 02:05:22 +0000
+++ bzrlib/fetch.py	2008-10-09 04:50:37 +0000
@@ -175,7 +175,9 @@
                     text_keys.extend([(file_id, revision) for revision in
                         revisions])
                 elif knit_kind == "inventory":
-                    # Now copy the file texts.
+                    # When the inventory keys start being reported, all text
+                    # keys have already been issued - and we want the text keys
+                    # inserted before inventory keys: copy the texts.
                     to_texts = self.to_repository.texts
                     from_texts = self.from_repository.texts
                     to_texts.insert_record_stream(from_texts.get_record_stream(
@@ -240,10 +242,19 @@
             # know for unselected inventories whether all their required
             # texts are present in the other repository - it could be
             # corrupt.
-            to_weave.insert_record_stream(from_weave.get_record_stream(
-                [(rev_id,) for rev_id in revs],
-                self.to_repository._fetch_order,
-                not self.to_repository._fetch_uses_deltas))
+            if self.to_repository._format.supports_chks:
+                # Hack to make chk->chk fetch: copy the inventories as
+                # inventories.
+                total = len(revs)
+                for pos, inv in enumerate(
+                    self.from_repository.iter_inventories(revs)):
+                    pb.update("Copying inventories", pos, total)
+                    self.to_repository.add_inventory(inv.revision_id, inv, [])
+            else:
+                to_weave.insert_record_stream(from_weave.get_record_stream(
+                    [(rev_id,) for rev_id in revs],
+                    self.to_repository._fetch_order,
+                    not self.to_repository._fetch_uses_deltas))
         finally:
             child_pb.finished()
 

=== modified file 'bzrlib/inventory.py'
--- bzrlib/inventory.py	2008-08-09 00:45:30 +0000
+++ bzrlib/inventory.py	2008-10-09 04:50:37 +0000
@@ -38,6 +38,7 @@
 
 import bzrlib
 from bzrlib import (
+    chk_map,
     errors,
     generate_ids,
     osutils,
@@ -90,16 +91,16 @@
     >>> i.add(InventoryDirectory('123', 'src', ROOT_ID))
     InventoryDirectory('123', 'src', parent_id='TREE_ROOT', revision=None)
     >>> i.add(InventoryFile('2323', 'hello.c', parent_id='123'))
-    InventoryFile('2323', 'hello.c', parent_id='123', sha1=None, len=None)
+    InventoryFile('2323', 'hello.c', parent_id='123', sha1=None, len=None, revision=None)
     >>> shouldbe = {0: '', 1: 'src', 2: 'src/hello.c'}
     >>> for ix, j in enumerate(i.iter_entries()):
     ...   print (j[0] == shouldbe[ix], j[1])
     ... 
     (True, InventoryDirectory('TREE_ROOT', u'', parent_id=None, revision=None))
     (True, InventoryDirectory('123', 'src', parent_id='TREE_ROOT', revision=None))
-    (True, InventoryFile('2323', 'hello.c', parent_id='123', sha1=None, len=None))
+    (True, InventoryFile('2323', 'hello.c', parent_id='123', sha1=None, len=None, revision=None))
     >>> i.add(InventoryFile('2324', 'bye.c', '123'))
-    InventoryFile('2324', 'bye.c', parent_id='123', sha1=None, len=None)
+    InventoryFile('2324', 'bye.c', parent_id='123', sha1=None, len=None, revision=None)
     >>> i.add(InventoryDirectory('2325', 'wibble', '123'))
     InventoryDirectory('2325', 'wibble', parent_id='123', revision=None)
     >>> i.path2id('src/wibble')
@@ -107,9 +108,9 @@
     >>> '2325' in i
     True
     >>> i.add(InventoryFile('2326', 'wibble.c', '2325'))
-    InventoryFile('2326', 'wibble.c', parent_id='2325', sha1=None, len=None)
+    InventoryFile('2326', 'wibble.c', parent_id='2325', sha1=None, len=None, revision=None)
     >>> i['2326']
-    InventoryFile('2326', 'wibble.c', parent_id='2325', sha1=None, len=None)
+    InventoryFile('2326', 'wibble.c', parent_id='2325', sha1=None, len=None, revision=None)
     >>> for path, entry in i.iter_entries():
     ...     print path
     ... 
@@ -564,13 +565,14 @@
         self.executable = work_tree.is_executable(self.file_id, path=path)
 
     def __repr__(self):
-        return ("%s(%r, %r, parent_id=%r, sha1=%r, len=%s)"
+        return ("%s(%r, %r, parent_id=%r, sha1=%r, len=%s, revision=%s)"
                 % (self.__class__.__name__,
                    self.file_id,
                    self.name,
                    self.parent_id,
                    self.text_sha1,
-                   self.text_size))
+                   self.text_size,
+                   self.revision))
 
     def _forget_tree_state(self):
         self.text_sha1 = None
@@ -709,7 +711,128 @@
         return compatible
 
 
-class Inventory(object):
+class CommonInventory(object):
+    """Basic inventory logic, defined in terms of primitives like has_id."""
+
+    def iter_entries(self, from_dir=None):
+        """Return (path, entry) pairs, in order by name."""
+        if from_dir is None:
+            if self.root is None:
+                return
+            from_dir = self.root
+            yield '', self.root
+        elif isinstance(from_dir, basestring):
+            from_dir = self[from_dir]
+            
+        # unrolling the recursive called changed the time from
+        # 440ms/663ms (inline/total) to 116ms/116ms
+        children = from_dir.children.items()
+        children.sort()
+        children = collections.deque(children)
+        stack = [(u'', children)]
+        while stack:
+            from_dir_relpath, children = stack[-1]
+
+            while children:
+                name, ie = children.popleft()
+
+                # we know that from_dir_relpath never ends in a slash
+                # and 'f' doesn't begin with one, we can do a string op, rather
+                # than the checks of pathjoin(), though this means that all paths
+                # start with a slash
+                path = from_dir_relpath + '/' + name
+
+                yield path[1:], ie
+
+                if ie.kind != 'directory':
+                    continue
+
+                # But do this child first
+                new_children = ie.children.items()
+                new_children.sort()
+                new_children = collections.deque(new_children)
+                stack.append((path, new_children))
+                # Break out of inner loop, so that we start outer loop with child
+                break
+            else:
+                # if we finished all children, pop it off the stack
+                stack.pop()
+
+    def iter_entries_by_dir(self, from_dir=None, specific_file_ids=None,
+        yield_parents=False):
+        """Iterate over the entries in a directory first order.
+
+        This returns all entries for a directory before returning
+        the entries for children of a directory. This is not
+        lexicographically sorted order, and is a hybrid between
+        depth-first and breadth-first.
+
+        :param yield_parents: If True, yield the parents from the root leading
+            down to specific_file_ids that have been requested. This has no
+            impact if specific_file_ids is None.
+        :return: This yields (path, entry) pairs
+        """
+        if specific_file_ids and not isinstance(specific_file_ids, set):
+            specific_file_ids = set(specific_file_ids)
+        # TODO? Perhaps this should return the from_dir so that the root is
+        # yielded? or maybe an option?
+        if from_dir is None:
+            if self.root is None:
+                return
+            # Optimize a common case
+            if (not yield_parents and specific_file_ids is not None and
+                len(specific_file_ids) == 1):
+                file_id = list(specific_file_ids)[0]
+                if file_id in self:
+                    yield self.id2path(file_id), self[file_id]
+                return 
+            from_dir = self.root
+            if (specific_file_ids is None or yield_parents or
+                self.root.file_id in specific_file_ids):
+                yield u'', self.root
+        elif isinstance(from_dir, basestring):
+            from_dir = self[from_dir]
+
+        if specific_file_ids is not None:
+            # TODO: jam 20070302 This could really be done as a loop rather
+            #       than a bunch of recursive calls.
+            parents = set()
+            byid = self
+            def add_ancestors(file_id):
+                if file_id not in byid:
+                    return
+                parent_id = byid[file_id].parent_id
+                if parent_id is None:
+                    return
+                if parent_id not in parents:
+                    parents.add(parent_id)
+                    add_ancestors(parent_id)
+            for file_id in specific_file_ids:
+                add_ancestors(file_id)
+        else:
+            parents = None
+            
+        stack = [(u'', from_dir)]
+        while stack:
+            cur_relpath, cur_dir = stack.pop()
+
+            child_dirs = []
+            for child_name, child_ie in sorted(cur_dir.children.iteritems()):
+
+                child_relpath = cur_relpath + child_name
+
+                if (specific_file_ids is None or 
+                    child_ie.file_id in specific_file_ids or
+                    (yield_parents and child_ie.file_id in parents)):
+                    yield child_relpath, child_ie
+
+                if child_ie.kind == 'directory':
+                    if parents is None or child_ie.file_id in parents:
+                        child_dirs.append((child_relpath+'/', child_ie))
+            stack.extend(reversed(child_dirs))
+
+
+class Inventory(CommonInventory):
     """Inventory of versioned files in a tree.
 
     This describes which file_id is present at each point in the tree,
@@ -728,7 +851,7 @@
 
     >>> inv = Inventory()
     >>> inv.add(InventoryFile('123-123', 'hello.c', ROOT_ID))
-    InventoryFile('123-123', 'hello.c', parent_id='TREE_ROOT', sha1=None, len=None)
+    InventoryFile('123-123', 'hello.c', parent_id='TREE_ROOT', sha1=None, len=None, revision=None)
     >>> inv['123-123'].name
     'hello.c'
 
@@ -748,7 +871,7 @@
     Traceback (most recent call last):
     BzrError: parent_id {TREE_ROOT} not in inventory
     >>> inv.add(InventoryFile('123-123', 'hello.c', 'TREE_ROOT-12345678-12345678'))
-    InventoryFile('123-123', 'hello.c', parent_id='TREE_ROOT-12345678-12345678', sha1=None, len=None)
+    InventoryFile('123-123', 'hello.c', parent_id='TREE_ROOT-12345678-12345678', sha1=None, len=None, revision=None)
     """
     def __init__(self, root_id=ROOT_ID, revision_id=None):
         """Create or read an inventory.
@@ -864,123 +987,6 @@
         """Returns number of entries."""
         return len(self._byid)
 
-    def iter_entries(self, from_dir=None):
-        """Return (path, entry) pairs, in order by name."""
-        if from_dir is None:
-            if self.root is None:
-                return
-            from_dir = self.root
-            yield '', self.root
-        elif isinstance(from_dir, basestring):
-            from_dir = self._byid[from_dir]
-            
-        # unrolling the recursive called changed the time from
-        # 440ms/663ms (inline/total) to 116ms/116ms
-        children = from_dir.children.items()
-        children.sort()
-        children = collections.deque(children)
-        stack = [(u'', children)]
-        while stack:
-            from_dir_relpath, children = stack[-1]
-
-            while children:
-                name, ie = children.popleft()
-
-                # we know that from_dir_relpath never ends in a slash
-                # and 'f' doesn't begin with one, we can do a string op, rather
-                # than the checks of pathjoin(), though this means that all paths
-                # start with a slash
-                path = from_dir_relpath + '/' + name
-
-                yield path[1:], ie
-
-                if ie.kind != 'directory':
-                    continue
-
-                # But do this child first
-                new_children = ie.children.items()
-                new_children.sort()
-                new_children = collections.deque(new_children)
-                stack.append((path, new_children))
-                # Break out of inner loop, so that we start outer loop with child
-                break
-            else:
-                # if we finished all children, pop it off the stack
-                stack.pop()
-
-    def iter_entries_by_dir(self, from_dir=None, specific_file_ids=None,
-        yield_parents=False):
-        """Iterate over the entries in a directory first order.
-
-        This returns all entries for a directory before returning
-        the entries for children of a directory. This is not
-        lexicographically sorted order, and is a hybrid between
-        depth-first and breadth-first.
-
-        :param yield_parents: If True, yield the parents from the root leading
-            down to specific_file_ids that have been requested. This has no
-            impact if specific_file_ids is None.
-        :return: This yields (path, entry) pairs
-        """
-        if specific_file_ids and not isinstance(specific_file_ids, set):
-            specific_file_ids = set(specific_file_ids)
-        # TODO? Perhaps this should return the from_dir so that the root is
-        # yielded? or maybe an option?
-        if from_dir is None:
-            if self.root is None:
-                return
-            # Optimize a common case
-            if (not yield_parents and specific_file_ids is not None and
-                len(specific_file_ids) == 1):
-                file_id = list(specific_file_ids)[0]
-                if file_id in self:
-                    yield self.id2path(file_id), self[file_id]
-                return 
-            from_dir = self.root
-            if (specific_file_ids is None or yield_parents or
-                self.root.file_id in specific_file_ids):
-                yield u'', self.root
-        elif isinstance(from_dir, basestring):
-            from_dir = self._byid[from_dir]
-
-        if specific_file_ids is not None:
-            # TODO: jam 20070302 This could really be done as a loop rather
-            #       than a bunch of recursive calls.
-            parents = set()
-            byid = self._byid
-            def add_ancestors(file_id):
-                if file_id not in byid:
-                    return
-                parent_id = byid[file_id].parent_id
-                if parent_id is None:
-                    return
-                if parent_id not in parents:
-                    parents.add(parent_id)
-                    add_ancestors(parent_id)
-            for file_id in specific_file_ids:
-                add_ancestors(file_id)
-        else:
-            parents = None
-            
-        stack = [(u'', from_dir)]
-        while stack:
-            cur_relpath, cur_dir = stack.pop()
-
-            child_dirs = []
-            for child_name, child_ie in sorted(cur_dir.children.iteritems()):
-
-                child_relpath = cur_relpath + child_name
-
-                if (specific_file_ids is None or 
-                    child_ie.file_id in specific_file_ids or
-                    (yield_parents and child_ie.file_id in parents)):
-                    yield child_relpath, child_ie
-
-                if child_ie.kind == 'directory':
-                    if parents is None or child_ie.file_id in parents:
-                        child_dirs.append((child_relpath+'/', child_ie))
-            stack.extend(reversed(child_dirs))
-
     def make_entry(self, kind, name, parent_id, file_id=None):
         """Simple thunk to bzrlib.inventory.make_entry."""
         return make_entry(kind, name, parent_id, file_id)
@@ -1024,7 +1030,7 @@
 
         >>> inv = Inventory()
         >>> inv.add(InventoryFile('123', 'foo.c', ROOT_ID))
-        InventoryFile('123', 'foo.c', parent_id='TREE_ROOT', sha1=None, len=None)
+        InventoryFile('123', 'foo.c', parent_id='TREE_ROOT', sha1=None, len=None, revision=None)
         >>> '123' in inv
         True
         >>> '456' in inv
@@ -1037,7 +1043,7 @@
 
         >>> inv = Inventory()
         >>> inv.add(InventoryFile('123123', 'hello.c', ROOT_ID))
-        InventoryFile('123123', 'hello.c', parent_id='TREE_ROOT', sha1=None, len=None)
+        InventoryFile('123123', 'hello.c', parent_id='TREE_ROOT', sha1=None, len=None, revision=None)
         >>> inv['123123'].name
         'hello.c'
         """
@@ -1119,7 +1125,7 @@
 
         >>> inv = Inventory()
         >>> inv.add(InventoryFile('123', 'foo.c', ROOT_ID))
-        InventoryFile('123', 'foo.c', parent_id='TREE_ROOT', sha1=None, len=None)
+        InventoryFile('123', 'foo.c', parent_id='TREE_ROOT', sha1=None, len=None, revision=None)
         >>> '123' in inv
         True
         >>> del inv['123']
@@ -1139,11 +1145,11 @@
         >>> i1 == i2
         True
         >>> i1.add(InventoryFile('123', 'foo', ROOT_ID))
-        InventoryFile('123', 'foo', parent_id='TREE_ROOT', sha1=None, len=None)
+        InventoryFile('123', 'foo', parent_id='TREE_ROOT', sha1=None, len=None, revision=None)
         >>> i1 == i2
         False
         >>> i2.add(InventoryFile('123', 'foo', ROOT_ID))
-        InventoryFile('123', 'foo', parent_id='TREE_ROOT', sha1=None, len=None)
+        InventoryFile('123', 'foo', parent_id='TREE_ROOT', sha1=None, len=None, revision=None)
         >>> i1 == i2
         True
         """
@@ -1288,6 +1294,199 @@
         return self.root is not None and file_id == self.root.file_id
 
 
+class CHKInventory(CommonInventory):
+    """A inventory persisted in a CHK store.
+    
+    No caching is performed: every method call or item access will perform
+    requests to the storage layer. As such, keep references to objects you want
+    to reuse.
+    """
+
+    def _entry_to_bytes(self, entry):
+        """Serialise entry as a single bytestring.
+
+        :param Entry: An inventory entry.
+        :return: A bytestring for the entry.
+        
+        The BNF:
+        ENTRY ::= FILE | DIR | SYMLINK | TREE
+        FILE ::= "file: " COMMON NULL SHA NULL SIZE NULL EXECUTABLE
+        DIR ::= "dir: " COMMON
+        SYMLINK ::= "symlink: " COMMON NULL TARGET_UTF8
+        TREE ::= "tree: " COMMON REFERENCE_REVISION
+        COMMON ::= FILE_ID NULL PARENT_ID NULL NAME_UTF8 NULL REVISION
+        """
+        if entry.parent_id is not None:
+            parent_str = entry.parent_id
+        else:
+            parent_str = ''
+        name_str = entry.name.encode("utf8")
+        if entry.kind == 'file':
+            if entry.executable:
+                exec_str = "Y"
+            else:
+                exec_str = "N"
+            return "file: %s\x00%s\x00%s\x00%s\x00%s\x00%d\x00%s" % (
+                entry.file_id, parent_str, name_str, entry.revision,
+                entry.text_sha1, entry.text_size, exec_str)
+        elif entry.kind == 'directory':
+            return "dir: %s\x00%s\x00%s\x00%s" % (
+                entry.file_id, parent_str, name_str, entry.revision)
+        elif entry.kind == 'symlink':
+            return "symlink: %s\x00%s\x00%s\x00%s\x00%s" % (
+                entry.file_id, parent_str, name_str, entry.revision,
+                entry.symlink_target.encode("utf8"))
+        elif entry.kind == 'tree-reference':
+            return "tree: %s\x00%s\x00%s\x00%s\x00%s" % (
+                entry.file_id, parent_str, name_str, entry.revision,
+                entry.reference_revision)
+        else:
+            raise ValueError("unknown kind %r" % entry.kind)
+
+    def _bytes_to_entry(self, bytes):
+        """Deserialise a serialised entry."""
+        sections = bytes.split('\x00')
+        if sections[0].startswith("file: "):
+            result = InventoryFile(sections[0][6:],
+                sections[2].decode('utf8'),
+                sections[1])
+            result.text_sha1 = sections[4]
+            result.text_size = int(sections[5])
+            result.executable = sections[6] == "Y"
+        elif sections[0].startswith("dir: "):
+            result = CHKInventoryDirectory(sections[0][5:],
+                sections[2].decode('utf8'),
+                sections[1], self)
+        elif sections[0].startswith("symlink: "):
+            result = InventoryLink(sections[0][9:],
+                sections[2].decode('utf8'),
+                sections[1])
+            result.symlink_target = sections[4]
+        elif sections[0].startswith("tree: "):
+            result = TreeReference(sections[0][6:],
+                sections[2].decode('utf8'),
+                sections[1])
+            result.reference_revision = sections[4]
+        else:
+            raise ValueError("Not a serialised entry %r" % bytes)
+        result.revision = sections[3]
+        if result.parent_id == '':
+            result.parent_id = None
+        return result
+
+    @classmethod
+    def deserialise(klass, chk_store, bytes, expected_revision_id):
+        """Deserialise a CHKInventory.
+
+        :param chk_store: A CHK capable VersionedFiles instance.
+        :param bytes: The serialised bytes.
+        :param expected_revision_id: The revision ID we think this inventory is
+            for.
+        :return: A CHKInventory
+        """
+        result = CHKInventory()
+        lines = bytes.splitlines()
+        if lines[0] != 'chkinventory:':
+            raise ValueError("not a serialised CHKInventory: %r" % bytes)
+        result.revision_id = lines[1][13:]
+        result.root_id = lines[2][9:]
+        result.id_to_entry = chk_map.CHKMap(chk_store, (lines[3][13:],))
+        if (result.revision_id,) != expected_revision_id:
+            raise ValueError("Mismatched revision id and expected: %r, %r" %
+                (result.revision_id, expected_revision_id))
+        return result
+
+    @classmethod
+    def from_inventory(klass, chk_store, inventory):
+        """Create a CHKInventory from an existing inventory.
+
+        The content of inventory is copied into the chk_store, and a
+        CHKInventory referencing that is returned.
+
+        :param chk_store: A CHK capable VersionedFiles instance.
+        :param inventory: The inventory to copy.
+        """
+        result = CHKInventory()
+        result.revision_id = inventory.revision_id
+        result.root_id = inventory.root.file_id
+        result.id_to_entry = chk_map.CHKMap(chk_store, None)
+        for path, entry in inventory.iter_entries():
+            result.id_to_entry._map(entry.file_id,
+                result._entry_to_bytes(entry))
+        result.id_to_entry._save()
+        return result
+
+    def __getitem__(self, file_id):
+        """map a single file_id -> InventoryEntry."""
+        try:
+            return self._bytes_to_entry(
+                self.id_to_entry.iteritems([file_id]).next()[1])
+        except StopIteration:
+            raise KeyError(file_id)
+
+    def has_id(self, file_id):
+        # Perhaps have an explicit 'contains' method on CHKMap ?
+        return len(list(self.id_to_entry.iteritems([file_id]))) == 1
+
+    def __iter__(self):
+        """Iterate over the entire inventory contents; size-of-tree - beware!."""
+        for file_id, _ in self.id_to_entry.iteritems():
+            yield file_id
+
+    def __len__(self):
+        """Return the number of entries in the inventory."""
+        # Might want to cache the length in the meta node.
+        return len([item for item in self])
+
+    def to_lines(self):
+        """Serialise the inventory to lines."""
+        lines = ["chkinventory:\n"]
+        lines.append("revision_id: %s\n" % self.revision_id)
+        lines.append("root_id: %s\n" % self.root_id)
+        lines.append("id_to_entry: %s\n" % self.id_to_entry._root_node._key)
+        return lines
+
+    @property
+    def root(self):
+        """Get the root entry."""
+        return self[self.root_id]
+
+
+class CHKInventoryDirectory(InventoryDirectory):
+    """A directory in an inventory."""
+
+    __slots__ = ['text_sha1', 'text_size', 'file_id', 'name', 'kind',
+                 'text_id', 'parent_id', '_children', 'executable',
+                 'revision', 'symlink_target', 'reference_revision',
+                 '_chk_inventory']
+
+    def __init__(self, file_id, name, parent_id, chk_inventory):
+        # Don't call InventoryDirectory.__init__ - it isn't right for this
+        # class.
+        InventoryEntry.__init__(self, file_id, name, parent_id)
+        self._children = None
+        self.kind = 'directory'
+        self._chk_inventory = chk_inventory
+
+    @property
+    def children(self):
+        """Access the list of children of this inventory.
+
+        Currently causes a full-load of all the children; a more sophisticated
+        proxy object is planned.
+        """
+        if self._children is not None:
+            return self._children
+        result = {}
+        # populate; todo: do by name
+        for file_id, bytes in self._chk_inventory.id_to_entry.iteritems():
+            entry = self._chk_inventory._bytes_to_entry(bytes)
+            if entry.parent_id == self.file_id:
+                result[entry.name] = entry
+        self._children = result
+        return result
+
+
 entry_factory = {
     'directory': InventoryDirectory,
     'file': InventoryFile,

=== modified file 'bzrlib/reconcile.py'
--- bzrlib/reconcile.py	2008-08-18 22:34:21 +0000
+++ bzrlib/reconcile.py	2008-10-09 04:50:37 +0000
@@ -505,9 +505,8 @@
             total_inventories = len(list(
                 collection.inventory_index.combined_index.iter_all_entries()))
             if len(all_revisions):
-                self._packer = repofmt.pack_repo.ReconcilePacker(
-                    collection, packs, ".reconcile", all_revisions)
-                new_pack = self._packer.pack(pb=self.pb)
+                new_pack =  self.repo._reconcile_pack(collection, packs,
+                    ".reconcile", all_revisions, self.pb)
                 if new_pack is not None:
                     self._discard_and_save(packs)
             else:

=== modified file 'bzrlib/repofmt/pack_repo.py'
--- bzrlib/repofmt/pack_repo.py	2008-09-29 06:37:31 +0000
+++ bzrlib/repofmt/pack_repo.py	2008-10-09 04:50:37 +0000
@@ -21,6 +21,7 @@
 import time
 
 from bzrlib import (
+    chk_map,
     debug,
     graph,
     pack,
@@ -31,6 +32,7 @@
     CombinedGraphIndex,
     GraphIndexPrefixAdapter,
     )
+from bzrlib.inventory import CHKInventory
 from bzrlib.knit import (
     KnitPlainFactory,
     KnitVersionedFiles,
@@ -42,6 +44,7 @@
 """)
 from bzrlib import (
     bzrdir,
+    chk_serializer,
     errors,
     lockable_files,
     lockdir,
@@ -770,19 +773,13 @@
         # NB XXX: how to check CHK references are present? perhaps by yielding
         # the items? How should that interact with stacked repos?
         if new_pack.chk_index is not None:
-            # XXX: Todo, recursive follow-pointers facility when fetching some
-            # revisions only.
-            chk_index_map = self._pack_collection._packs_list_to_pack_map_and_index_list(
-                self.packs, 'chk_index')[0]
-            chk_nodes = self._pack_collection._index_contents(chk_index_map, None)
-            self._copy_nodes(chk_nodes, chk_index_map, new_pack._writer,
-                new_pack.chk_index)
-            if 'pack' in debug.debug_flags:
-                mutter('%s: create_pack: chk content copied: %s%s %d items t+%6.3fs',
-                    time.ctime(), self._pack_collection._upload_transport.base,
-                    new_pack.random_name,
-                    new_pack.chk_index.key_count(),
-                    time.time() - new_pack.start_time)
+            self._copy_chks()
+        if 'pack' in debug.debug_flags:
+            mutter('%s: create_pack: chk content copied: %s%s %d items t+%6.3fs',
+                time.ctime(), self._pack_collection._upload_transport.base,
+                new_pack.random_name,
+                new_pack.chk_index.key_count(),
+                time.time() - new_pack.start_time)
         self._check_references()
         if not self._use_pack(new_pack):
             new_pack.abort()
@@ -792,16 +789,43 @@
         self._pack_collection.allocate(new_pack)
         return new_pack
 
-    def _copy_nodes(self, nodes, index_map, writer, write_index):
-        """Copy knit nodes between packs with no graph references."""
+    def _copy_chks(self, refs=None):
+        # XXX: Todo, recursive follow-pointers facility when fetching some
+        # revisions only.
+        chk_index_map = self._pack_collection._packs_list_to_pack_map_and_index_list(
+            self.packs, 'chk_index')[0]
+        chk_nodes = self._pack_collection._index_contents(chk_index_map, refs)
+        new_refs = set()
+        def accumlate_refs(lines):
+            # XXX: move to a generic location
+            kind = lines[0]
+            if kind == 'chkroot:\n':
+                node = chk_map.RootNode()
+                node.deserialise(''.join(lines),  None)
+                new_refs.update(node.refs())
+            elif kind == 'chkvalue:\n':
+                pass
+            else:
+                raise ValueError("unknown chk node %r" % lines[0])
+        self._copy_nodes(chk_nodes, chk_index_map, self.new_pack._writer,
+            self.new_pack.chk_index, output_lines=accumlate_refs)
+        return new_refs
+
+    def _copy_nodes(self, nodes, index_map, writer, write_index,
+        output_lines=None):
+        """Copy knit nodes between packs with no graph references.
+        
+        :param output_lines: Output full texts of copied items.
+        """
         pb = ui.ui_factory.nested_progress_bar()
         try:
             return self._do_copy_nodes(nodes, index_map, writer,
-                write_index, pb)
+                write_index, pb, output_lines=output_lines)
         finally:
             pb.finished()
 
-    def _do_copy_nodes(self, nodes, index_map, writer, write_index, pb):
+    def _do_copy_nodes(self, nodes, index_map, writer, write_index, pb,
+        output_lines=None):
         # for record verification
         knit = KnitVersionedFiles(None, None)
         # plan a readv on each source pack:
@@ -835,8 +859,11 @@
                 izip(reader.iter_records(), pack_readv_requests):
                 raw_data = read_func(None)
                 # check the header only
-                df, _ = knit._parse_record_header(key, raw_data)
-                df.close()
+                if output_lines is not None:
+                    output_lines(knit._parse_record(key[-1], raw_data)[0])
+                else:
+                    df, _ = knit._parse_record_header(key, raw_data)
+                    df.close()
                 pos, size = writer.add_bytes_record(raw_data, names)
                 write_index.add_node(key, eol_flag + "%d %d" % (pos, size))
                 pb.update("Copied record", record_index)
@@ -1956,6 +1983,10 @@
         reconciler.reconcile()
         return reconciler
 
+    def _reconcile_pack(self, collection, packs, extension, revs, pb):
+        packer = ReconcilePacker(collection, packs, extension, revs)
+        return packer.pack(pb)
+
     def unlock(self):
         if self._write_lock_count == 1 and self._write_group is not None:
             self.abort_write_group()
@@ -1978,6 +2009,131 @@
                 repo.unlock()
 
 
+class CHKInventoryRepository(KnitPackRepository):
+    """subclass of KnitPackRepository that uses CHK based inventories."""
+
+    def _add_inventory_checked(self, revision_id, inv, parents):
+        """Add inv to the repository after checking the inputs.
+
+        This function can be overridden to allow different inventory styles.
+
+        :seealso: add_inventory, for the contract.
+        """
+        # make inventory
+        result = CHKInventory.from_inventory(self.chk_bytes, inv)
+        inv_lines = result.to_lines()
+        return self._inventory_add_lines(revision_id, parents,
+            inv_lines, check_content=False)
+
+    def _iter_inventories(self, revision_ids):
+        """Iterate over many inventory objects."""
+        keys = [(revision_id,) for revision_id in revision_ids]
+        stream = self.inventories.get_record_stream(keys, 'unordered', True)
+        texts = {}
+        for record in stream:
+            if record.storage_kind != 'absent':
+                texts[record.key] = record.get_bytes_as('fulltext')
+            else:
+                raise errors.NoSuchRevision(self, record.key)
+        for key in keys:
+            yield CHKInventory.deserialise(self.chk_bytes, texts[key], key)
+
+    def _iter_inventory_xmls(self, revision_ids):
+        # Without a native 'xml' inventory, this method doesn't make sense, so
+        # make it raise to trap naughty direct users.
+        raise NotImplementedError(self._iter_inventory_xmls)
+
+    def fileids_altered_by_revision_ids(self, revision_ids, _inv_weave=None):
+        """Find the file ids and versions affected by revisions.
+
+        :param revisions: an iterable containing revision ids.
+        :param _inv_weave: The inventory weave from this repository or None.
+            If None, the inventory weave will be opened automatically.
+        :return: a dictionary mapping altered file-ids to an iterable of
+            revision_ids. Each altered file-ids has the exact revision_ids that
+            altered it listed explicitly.
+        """
+        rich_roots = self.supports_rich_root()
+        result = {}
+        pb = ui.ui_factory.nested_progress_bar()
+        try:
+            total = len(revision_ids)
+            for pos, inv in enumerate(self.iter_inventories(revision_ids)):
+                pb.update("Finding text references", pos, total)
+                for _, entry in inv.iter_entries():
+                    if entry.revision != inv.revision_id:
+                        continue
+                    if not rich_roots and entry.file_id == inv.root_id:
+                        continue
+                    alterations = result.setdefault(entry.file_id, set([]))
+                    alterations.add(entry.revision)
+            return result
+        finally:
+            pb.finished()
+
+    def find_text_key_references(self):
+        """Find the text key references within the repository.
+
+        :return: A dictionary mapping text keys ((fileid, revision_id) tuples)
+            to whether they were referred to by the inventory of the
+            revision_id that they contain. The inventory texts from all present
+            revision ids are assessed to generate this report.
+        """
+        # XXX: Slow version but correct: rewrite as a series of delta
+        # examinations/direct tree traversal. Note that that will require care
+        # as a common node is reachable both from the inventory that added it,
+        # and others afterwards.
+        revision_keys = self.revisions.keys()
+        result = {}
+        rich_roots = self.supports_rich_root()
+        pb = ui.ui_factory.nested_progress_bar()
+        try:
+            all_revs = self.all_revision_ids()
+            total = len(all_revs)
+            for pos, inv in enumerate(self.iter_inventories(all_revs)):
+                pb.update("Finding text references", pos, total)
+                for _, entry in inv.iter_entries():
+                    if not rich_roots and entry.file_id == inv.root_id:
+                        continue
+                    key = (entry.file_id, entry.revision)
+                    result.setdefault(key, False)
+                    if entry.revision == inv.revision_id:
+                        result[key] = True
+            return result
+        finally:
+            pb.finished()
+
+    def _reconcile_pack(self, collection, packs, extension, revs, pb):
+        packer = CHKReconcilePacker(collection, packs, extension, revs)
+        return packer.pack(pb)
+
+
+class CHKReconcilePacker(ReconcilePacker):
+    """Subclass of ReconcilePacker for handling chk inventories."""
+
+    def _process_inventory_lines(self, inv_lines):
+        """Generate a text key reference map rather for reconciling with."""
+        repo = self._pack_collection.repo
+        # XXX: This double-reads the inventories; but it works.
+        refs = repo.find_text_key_references()
+        self._text_refs = refs
+        # during reconcile we:
+        #  - convert unreferenced texts to full texts
+        #  - correct texts which reference a text not copied to be full texts
+        #  - copy all others as-is but with corrected parents.
+        #  - so at this point we don't know enough to decide what becomes a full
+        #    text.
+        self._text_filter = None
+        # Copy the selected inventory roots, extracting the CHK references
+        # needed.
+        pending_refs = set()
+        for line, revid in inv_lines:
+            if line.startswith('id_to_entry: '):
+                pending_refs.add((line[13:],))
+        while pending_refs:
+            pending_refs = self._copy_chks(pending_refs)
+
+
 class RepositoryFormatPack(MetaDirRepositoryFormat):
     """Format logic for pack structured repositories.
 
@@ -2386,9 +2542,9 @@
     This is pack-1.6.1 with B+Tree indices and a chk index.
     """
 
-    repository_class = KnitPackRepository
+    repository_class = CHKInventoryRepository
     _commit_builder_class = PackCommitBuilder
-    _serializer = xml5.serializer_v5
+    _serializer = chk_serializer.chk_serializer
     supports_external_lookups = True
     # What index classes to use
     index_builder_class = BTreeBuilder
@@ -2424,11 +2580,11 @@
     1.6.1-subtree[as it might have been] with B+Tree indices and a chk index.
     """
 
-    repository_class = KnitPackRepository
+    repository_class = CHKInventoryRepository
     _commit_builder_class = PackRootCommitBuilder
     rich_root_data = True
     supports_tree_reference = True
-    _serializer = xml7.serializer_v7
+    _serializer = chk_serializer.chk_serializer_subtree
     supports_external_lookups = True
     # What index classes to use
     index_builder_class = BTreeBuilder

=== modified file 'bzrlib/repository.py'
--- bzrlib/repository.py	2008-09-26 02:00:48 +0000
+++ bzrlib/repository.py	2008-10-09 04:50:37 +0000
@@ -519,7 +519,9 @@
         """
         if self._write_group is not self.get_transaction():
             # has an unlock or relock occured ?
-            raise errors.BzrError('mismatched lock context and write group.')
+            raise errors.BzrError(
+                'mismatched lock context and write group. %r, %r' %
+                (self._write_group, self.get_transaction()))
         self._abort_write_group()
         self._write_group = None
 
@@ -549,6 +551,8 @@
         self.inventories.add_fallback_versioned_files(repository.inventories)
         self.revisions.add_fallback_versioned_files(repository.revisions)
         self.signatures.add_fallback_versioned_files(repository.signatures)
+        if self.chk_bytes is not None:
+            self.chk_bytes.add_fallback_versioned_files(repository.chk_bytes)
 
     def _check_fallback_repository(self, repository):
         """Check that this repository can fallback to repository safely.
@@ -578,6 +582,15 @@
                 % (inv.revision_id, revision_id))
         if inv.root is None:
             raise AssertionError()
+        return self._add_inventory_checked(revision_id, inv, parents)
+
+    def _add_inventory_checked(self, revision_id, inv, parents):
+        """Add inv to the repository after checking the inputs.
+
+        This function can be overridden to allow different inventory styles.
+
+        :seealso: add_inventory, for the contract.
+        """
         inv_lines = self._serialise_inventory_to_lines(inv)
         return self._inventory_add_lines(revision_id, parents,
             inv_lines, check_content=False)
@@ -1201,9 +1214,6 @@
     def find_text_key_references(self):
         """Find the text key references within the repository.
 
-        :return: a dictionary mapping (file_id, revision_id) tuples to altered file-ids to an iterable of
-        revision_ids. Each altered file-ids has the exact revision_ids that
-        altered it listed explicitly.
         :return: A dictionary mapping text keys ((fileid, revision_id) tuples)
             to whether they were referred to by the inventory of the
             revision_id that they contain. The inventory texts from all present
@@ -1458,7 +1468,10 @@
                             except KeyError:
                                 inv = self.revision_tree(parent_id).inventory
                                 inventory_cache[parent_id] = inv
-                            parent_entry = inv._byid.get(text_key[0], None)
+                            try:
+                                parent_entry = inv[text_key[0]]
+                            except (KeyError, errors.NoSuchId):
+                                parent_entry = None
                             if parent_entry is not None:
                                 parent_text_key = (
                                     text_key[0], parent_entry.revision)
@@ -2760,6 +2773,8 @@
         We don't test for the stores being of specific types because that
         could lead to confusing results, and there is no need to be 
         overly general.
+
+        Do not support CHK based repositories at this point.
         """
         from bzrlib.repofmt.pack_repo import RepositoryFormatPack
         try:
@@ -2767,7 +2782,8 @@
                 isinstance(target._format, RepositoryFormatPack))
         except AttributeError:
             return False
-        return are_packs and InterRepository._same_model(source, target)
+        return (are_packs and InterRepository._same_model(source, target) and
+            not source._format.supports_chks)
 
     @needs_write_lock
     def fetch(self, revision_id=None, pb=None, find_ghosts=False):

=== modified file 'bzrlib/tests/per_repository/test_check.py'
--- bzrlib/tests/per_repository/test_check.py	2008-09-04 20:32:04 +0000
+++ bzrlib/tests/per_repository/test_check.py	2008-10-09 04:50:37 +0000
@@ -43,7 +43,7 @@
         # XXX: check requires a non-empty revision IDs list, but it ignores the
         # contents of it!
         check_object = tree.branch.repository.check(['ignored'])
-        check_object.report_results(verbose=False)
+        check_object.report_results(verbose=True)
         log = self._get_log(keep_log_file=True)
         self.assertContainsRe(
             log,

=== modified file 'bzrlib/tests/per_repository/test_commit_builder.py'
--- bzrlib/tests/per_repository/test_commit_builder.py	2008-09-04 20:32:04 +0000
+++ bzrlib/tests/per_repository/test_commit_builder.py	2008-10-09 04:50:37 +0000
@@ -411,12 +411,13 @@
             self.assertEqual(expected_delta, delta)
             builder.finish_inventory()
             rev2 = builder.commit('')
+        except:
+            builder.abort()
+            tree.unlock()
+            raise
+        try:
             tree.set_parent_ids([rev2])
-        except:
-            builder.abort()
-            tree.unlock()
-            raise
-        else:
+        finally:
             tree.unlock()
         return rev2
 

=== modified file 'bzrlib/tests/per_repository/test_repository.py'
--- bzrlib/tests/per_repository/test_repository.py	2008-09-17 07:18:20 +0000
+++ bzrlib/tests/per_repository/test_repository.py	2008-10-09 04:50:37 +0000
@@ -29,7 +29,7 @@
     repository,
     )
 from bzrlib.delta import TreeDelta
-from bzrlib.inventory import Inventory, InventoryDirectory
+from bzrlib.inventory import CommonInventory, Inventory, InventoryDirectory
 from bzrlib.repofmt.weaverepo import (
     RepositoryFormat5,
     RepositoryFormat6,
@@ -242,13 +242,16 @@
         invs = tree.branch.repository.iter_inventories(revs)
         for rev_id, inv in zip(revs, invs):
             self.assertEqual(rev_id, inv.revision_id)
-            self.assertIsInstance(inv, Inventory)
+            self.assertIsInstance(inv, CommonInventory)
 
     def test_supports_rich_root(self):
         tree = self.make_branch_and_tree('a')
         tree.commit('')
         second_revision = tree.commit('')
-        inv = tree.branch.repository.revision_tree(second_revision).inventory
+        rev_tree = tree.branch.repository.revision_tree(second_revision)
+        rev_tree.lock_read()
+        self.addCleanup(rev_tree.unlock)
+        inv = rev_tree.inventory
         rich_root = (inv.root.revision != second_revision)
         self.assertEqual(rich_root,
                          tree.branch.repository.supports_rich_root())
@@ -330,17 +333,29 @@
         wt.set_root_id('fixed-root')
         wt.commit('lala!', rev_id='revision-1', allow_pointless=True)
         tree = wt.branch.repository.revision_tree('revision-1')
-        self.assertEqual('revision-1', tree.inventory.root.revision) 
-        expected = InventoryDirectory('fixed-root', '', None)
-        expected.revision = 'revision-1'
-        self.assertEqual([('', 'V', 'directory', 'fixed-root', expected)],
-                         list(tree.list_files(include_root=True)))
+        tree.lock_read()
+        try:
+            self.assertEqual('revision-1', tree.inventory.root.revision)
+            expected = InventoryDirectory('fixed-root', '', None)
+            expected.revision = 'revision-1'
+            self.assertEqual([('', 'V', 'directory', 'fixed-root', expected)],
+                             list(tree.list_files(include_root=True)))
+        finally:
+            tree.unlock()
         tree = self.callDeprecated(['NULL_REVISION should be used for the null'
             ' revision instead of None, as of bzr 0.91.'],
             wt.branch.repository.revision_tree, None)
-        self.assertEqual([], list(tree.list_files(include_root=True)))
+        tree.lock_read()
+        try:
+            self.assertEqual([], list(tree.list_files(include_root=True)))
+        finally:
+            tree.unlock()
         tree = wt.branch.repository.revision_tree(NULL_REVISION)
-        self.assertEqual([], list(tree.list_files(include_root=True)))
+        tree.lock_read()
+        try:
+            self.assertEqual([], list(tree.list_files(include_root=True)))
+        finally:
+            tree.unlock()
 
     def test_get_revision_delta(self):
         tree_a = self.make_branch_and_tree('a')
@@ -487,6 +502,8 @@
         tree = self.make_branch_and_tree('.')
         tree.commit('message', rev_id='rev_id')
         rev_tree = tree.branch.repository.revision_tree(tree.last_revision())
+        rev_tree.lock_read()
+        self.addCleanup(rev_tree.unlock)
         self.assertEqual('rev_id', rev_tree.inventory.root.revision)
 
     def test_upgrade_from_format4(self):
@@ -655,10 +672,20 @@
                          repo.get_signature_text('A'))
 
     def test_add_revision_inventory_sha1(self):
-        repo = self.make_repository('repo')
         inv = Inventory(revision_id='A')
         inv.root.revision = 'A'
         inv.root.file_id = 'fixed-root'
+        # Insert the inventory on its own to an identical repository, to get
+        # its sha1.
+        reference_repo = self.make_repository('reference_repo')
+        reference_repo.lock_write()
+        reference_repo.start_write_group()
+        inv_sha1 = reference_repo.add_inventory('A', inv, [])
+        reference_repo.abort_write_group()
+        reference_repo.unlock()
+        # Now insert a revision with this inventory, and it should get the same
+        # sha1.
+        repo = self.make_repository('repo')
         repo.lock_write()
         repo.start_write_group()
         repo.add_revision('A', Revision('A', committer='B', timestamp=0,
@@ -666,9 +693,7 @@
         repo.commit_write_group()
         repo.unlock()
         repo.lock_read()
-        self.assertEquals(osutils.sha_string(
-            repo._serializer.write_inventory_to_string(inv)),
-            repo.get_revision('A').inventory_sha1)
+        self.assertEquals(inv_sha1, repo.get_revision('A').inventory_sha1)
         repo.unlock()
 
     def test_install_revisions(self):

=== modified file 'bzrlib/tests/per_repository_chk/test_supported.py'
--- bzrlib/tests/per_repository_chk/test_supported.py	2008-09-29 06:37:31 +0000
+++ bzrlib/tests/per_repository_chk/test_supported.py	2008-10-09 04:50:37 +0000
@@ -70,7 +70,7 @@
         try:
             repo.start_write_group()
             try:
-                repo.chk_bytes.add_lines((None,), None, ["foo\n", "bar\n"],
+                repo.chk_bytes.add_lines((None,), None, ["chkroot:\n"],
                     random_id=True)
             except:
                 repo.abort_write_group()
@@ -79,7 +79,7 @@
                 repo.commit_write_group()
             repo.start_write_group()
             try:
-                repo.chk_bytes.add_lines((None,), None, ["foo\n"],
+                repo.chk_bytes.add_lines((None,), None, ["chkvalue:\n"],
                     random_id=True)
             except:
                 repo.abort_write_group()
@@ -88,8 +88,8 @@
                 repo.commit_write_group()
             repo.pack()
             self.assertEqual(
-                set([('sha1:4e48e2c9a3d2ca8a708cb0cc545700544efb5021',),
-                    ('sha1:f1d2d2f924e986ac86fdf7b36c94bcdf32beec15',),]),
+                set([('sha1:062ef095245d8617d7671e50a71b529d13d93022',),
+                    ('sha1:572d8da882e1ebf0f50f1e2da2d7a9cadadf4db5',)]),
                 repo.chk_bytes.keys())
         finally:
             repo.unlock()
@@ -98,8 +98,8 @@
         repo.lock_read()
         try:
             self.assertEqual(
-                set([('sha1:4e48e2c9a3d2ca8a708cb0cc545700544efb5021',),
-                    ('sha1:f1d2d2f924e986ac86fdf7b36c94bcdf32beec15',),]),
+                set([('sha1:062ef095245d8617d7671e50a71b529d13d93022',),
+                    ('sha1:572d8da882e1ebf0f50f1e2da2d7a9cadadf4db5',)]),
                 repo.chk_bytes.keys())
         finally:
             repo.unlock()

=== modified file 'bzrlib/tests/per_repository_reference/test_add_inventory.py'
--- bzrlib/tests/per_repository_reference/test_add_inventory.py	2008-06-12 04:15:26 +0000
+++ bzrlib/tests/per_repository_reference/test_add_inventory.py	2008-10-09 04:50:37 +0000
@@ -30,6 +30,8 @@
         tree = self.make_branch_and_tree('sample')
         revid = tree.commit('one')
         inv = tree.branch.repository.get_inventory(revid)
+        tree.lock_read()
+        self.addCleanup(tree.unlock)
         base = self.make_repository('base')
         repo = self.make_referring('referring', 'base')
         repo.lock_write()
@@ -44,6 +46,10 @@
                 repo.commit_write_group()
         finally:
             repo.unlock()
+        repo.lock_read()
+        self.addCleanup(repo.unlock)
         inv2 = repo.get_inventory(revid)
-        self.assertEqual(inv._byid, inv2._byid)
+        content1 = dict((file_id, inv[file_id]) for file_id in inv)
+        content2 = dict((file_id, inv[file_id]) for file_id in inv2)
+        self.assertEqual(content1, content2)
         self.assertRaises(errors.NoSuchRevision, base.get_inventory, revid)

=== modified file 'bzrlib/tests/per_repository_reference/test_add_revision.py'
--- bzrlib/tests/per_repository_reference/test_add_revision.py	2008-06-06 08:20:03 +0000
+++ bzrlib/tests/per_repository_reference/test_add_revision.py	2008-10-09 04:50:37 +0000
@@ -30,6 +30,8 @@
         tree = self.make_branch_and_tree('sample')
         revid = tree.commit('one')
         inv = tree.branch.repository.get_inventory(revid)
+        tree.lock_read()
+        self.addCleanup(tree.unlock)
         rev = tree.branch.repository.get_revision(revid)
         base = self.make_repository('base')
         repo = self.make_referring('referring', 'base')

=== modified file 'bzrlib/tests/per_repository_reference/test_add_signature_text.py'
--- bzrlib/tests/per_repository_reference/test_add_signature_text.py	2008-06-06 08:20:03 +0000
+++ bzrlib/tests/per_repository_reference/test_add_signature_text.py	2008-10-09 04:50:37 +0000
@@ -30,6 +30,8 @@
         tree = self.make_branch_and_tree('sample')
         revid = tree.commit('one')
         inv = tree.branch.repository.get_inventory(revid)
+        tree.lock_read()
+        self.addCleanup(tree.unlock)
         base = self.make_repository('base')
         repo = self.make_referring('referring', 'base')
         repo.lock_write()

=== modified file 'bzrlib/tests/test_chk_map.py'
--- bzrlib/tests/test_chk_map.py	2008-10-01 03:46:28 +0000
+++ bzrlib/tests/test_chk_map.py	2008-10-09 04:50:37 +0000
@@ -106,6 +106,14 @@
         self.assertEqual([("a", "content here"), ("b", "more content")],
             sorted(list(chkmap.iteritems())))
 
+    def test_iteritems_selected_one_of_two_items(self):
+        chk_bytes = self.get_chk_bytes()
+        root_key = CHKMap.from_dict(chk_bytes,
+            {"a":"content here", "b":"more content"})
+        chkmap = CHKMap(chk_bytes, root_key)
+        self.assertEqual([("a", "content here")],
+            sorted(list(chkmap.iteritems(["a"]))))
+
 
 class TestRootNode(TestCaseWithTransport):
 

=== modified file 'bzrlib/tests/test_inv.py'
--- bzrlib/tests/test_inv.py	2007-08-21 05:29:59 +0000
+++ bzrlib/tests/test_inv.py	2008-10-09 04:50:37 +0000
@@ -16,9 +16,9 @@
 
 
 from bzrlib import errors, inventory, osutils
-from bzrlib.inventory import (Inventory, ROOT_ID, InventoryFile,
+from bzrlib.inventory import (CHKInventory, Inventory, ROOT_ID, InventoryFile,
     InventoryDirectory, InventoryEntry, TreeReference)
-from bzrlib.tests import TestCase
+from bzrlib.tests import TestCase, TestCaseWithTransport
 
 
 class TestInventoryEntry(TestCase):
@@ -166,3 +166,167 @@
     def assertChangeDescription(self, expected_change, old_ie, new_ie):
         change = InventoryEntry.describe_change(old_ie, new_ie)
         self.assertEqual(expected_change, change)
+
+
+class TestCHKInventory(TestCaseWithTransport):
+    
+    def get_chk_bytes(self):
+        # The eassiest way to get a CHK store is a development3 repository and
+        # then work with the chk_bytes attribute directly.
+        repo = self.make_repository(".", format="development3")
+        repo.lock_write()
+        self.addCleanup(repo.unlock)
+        repo.start_write_group()
+        self.addCleanup(repo.abort_write_group)
+        return repo.chk_bytes
+
+    def read_bytes(self, chk_bytes, key):
+        stream = chk_bytes.get_record_stream([key], 'unordered', True)
+        return stream.next().get_bytes_as("fulltext")
+
+    def test_deserialise_gives_CHKInventory(self):
+        inv = Inventory()
+        inv.revision_id = "revid"
+        inv.root.revision = "rootrev"
+        chk_bytes = self.get_chk_bytes()
+        chk_inv = CHKInventory.from_inventory(chk_bytes, inv)
+        bytes = ''.join(chk_inv.to_lines())
+        new_inv = CHKInventory.deserialise(chk_bytes, bytes, ("revid",))
+        self.assertEqual("revid", new_inv.revision_id)
+        self.assertEqual("directory", new_inv.root.kind)
+        self.assertEqual(inv.root.file_id, new_inv.root.file_id)
+        self.assertEqual(inv.root.parent_id, new_inv.root.parent_id)
+        self.assertEqual(inv.root.name, new_inv.root.name)
+        self.assertEqual("rootrev", new_inv.root.revision)
+
+    def test_deserialise_wrong_revid(self):
+        inv = Inventory()
+        inv.revision_id = "revid"
+        inv.root.revision = "rootrev"
+        chk_bytes = self.get_chk_bytes()
+        chk_inv = CHKInventory.from_inventory(chk_bytes, inv)
+        bytes = ''.join(chk_inv.to_lines())
+        self.assertRaises(ValueError, CHKInventory.deserialise, chk_bytes,
+            bytes, ("revid2",))
+
+    def test_captures_rev_root_byid(self):
+        inv = Inventory()
+        inv.revision_id = "foo"
+        inv.root.revision = "bar"
+        chk_bytes = self.get_chk_bytes()
+        chk_inv = CHKInventory.from_inventory(chk_bytes, inv)
+        self.assertEqual([
+            'chkinventory:\n',
+            'revision_id: foo\n',
+            'root_id: TREE_ROOT\n',
+            'id_to_entry: sha1:dc696b0cf291ac0d66bdcda3070f755494a586fc\n'
+            ],
+            chk_inv.to_lines())
+
+    def test_directory_children_on_demand(self):
+        inv = Inventory()
+        inv.revision_id = "revid"
+        inv.root.revision = "rootrev"
+        inv.add(InventoryFile("fileid", "file", inv.root.file_id))
+        inv["fileid"].revision = "filerev"
+        inv["fileid"].executable = True
+        inv["fileid"].text_sha1 = "ffff"
+        inv["fileid"].text_size = 1
+        chk_bytes = self.get_chk_bytes()
+        chk_inv = CHKInventory.from_inventory(chk_bytes, inv)
+        bytes = ''.join(chk_inv.to_lines())
+        new_inv = CHKInventory.deserialise(chk_bytes, bytes, ("revid",))
+        root_entry = new_inv[inv.root.file_id]
+        self.assertEqual(None, root_entry._children)
+        self.assertEqual(['file'], root_entry.children.keys())
+        file_direct = new_inv["fileid"]
+        file_found = root_entry.children['file']
+        self.assertEqual(file_direct.kind, file_found.kind)
+        self.assertEqual(file_direct.file_id, file_found.file_id)
+        self.assertEqual(file_direct.parent_id, file_found.parent_id)
+        self.assertEqual(file_direct.name, file_found.name)
+        self.assertEqual(file_direct.revision, file_found.revision)
+        self.assertEqual(file_direct.text_sha1, file_found.text_sha1)
+        self.assertEqual(file_direct.text_size, file_found.text_size)
+        self.assertEqual(file_direct.executable, file_found.executable)
+
+    def test___iter__(self):
+        inv = Inventory()
+        inv.revision_id = "revid"
+        inv.root.revision = "rootrev"
+        inv.add(InventoryFile("fileid", "file", inv.root.file_id))
+        inv["fileid"].revision = "filerev"
+        inv["fileid"].executable = True
+        inv["fileid"].text_sha1 = "ffff"
+        inv["fileid"].text_size = 1
+        chk_bytes = self.get_chk_bytes()
+        chk_inv = CHKInventory.from_inventory(chk_bytes, inv)
+        bytes = ''.join(chk_inv.to_lines())
+        new_inv = CHKInventory.deserialise(chk_bytes, bytes, ("revid",))
+        fileids = list(new_inv.__iter__())
+        fileids.sort()
+        self.assertEqual([inv.root.file_id, "fileid"], fileids)
+
+    def test__len__(self):
+        inv = Inventory()
+        inv.revision_id = "revid"
+        inv.root.revision = "rootrev"
+        inv.add(InventoryFile("fileid", "file", inv.root.file_id))
+        inv["fileid"].revision = "filerev"
+        inv["fileid"].executable = True
+        inv["fileid"].text_sha1 = "ffff"
+        inv["fileid"].text_size = 1
+        chk_bytes = self.get_chk_bytes()
+        chk_inv = CHKInventory.from_inventory(chk_bytes, inv)
+        self.assertEqual(2, len(chk_inv))
+
+    def test___getitem__(self):
+        inv = Inventory()
+        inv.revision_id = "revid"
+        inv.root.revision = "rootrev"
+        inv.add(InventoryFile("fileid", "file", inv.root.file_id))
+        inv["fileid"].revision = "filerev"
+        inv["fileid"].executable = True
+        inv["fileid"].text_sha1 = "ffff"
+        inv["fileid"].text_size = 1
+        chk_bytes = self.get_chk_bytes()
+        chk_inv = CHKInventory.from_inventory(chk_bytes, inv)
+        bytes = ''.join(chk_inv.to_lines())
+        new_inv = CHKInventory.deserialise(chk_bytes, bytes, ("revid",))
+        root_entry = new_inv[inv.root.file_id]
+        file_entry = new_inv["fileid"]
+        self.assertEqual("directory", root_entry.kind)
+        self.assertEqual(inv.root.file_id, root_entry.file_id)
+        self.assertEqual(inv.root.parent_id, root_entry.parent_id)
+        self.assertEqual(inv.root.name, root_entry.name)
+        self.assertEqual("rootrev", root_entry.revision)
+        self.assertEqual("file", file_entry.kind)
+        self.assertEqual("fileid", file_entry.file_id)
+        self.assertEqual(inv.root.file_id, file_entry.parent_id)
+        self.assertEqual("file", file_entry.name)
+        self.assertEqual("filerev", file_entry.revision)
+        self.assertEqual("ffff", file_entry.text_sha1)
+        self.assertEqual(1, file_entry.text_size)
+        self.assertEqual(True, file_entry.executable)
+
+    def test_has_id_true(self):
+        inv = Inventory()
+        inv.revision_id = "revid"
+        inv.root.revision = "rootrev"
+        inv.add(InventoryFile("fileid", "file", inv.root.file_id))
+        inv["fileid"].revision = "filerev"
+        inv["fileid"].executable = True
+        inv["fileid"].text_sha1 = "ffff"
+        inv["fileid"].text_size = 1
+        chk_bytes = self.get_chk_bytes()
+        chk_inv = CHKInventory.from_inventory(chk_bytes, inv)
+        self.assertTrue(chk_inv.has_id('fileid'))
+        self.assertTrue(chk_inv.has_id(inv.root.file_id))
+
+    def test_has_id_not(self):
+        inv = Inventory()
+        inv.revision_id = "revid"
+        inv.root.revision = "rootrev"
+        chk_bytes = self.get_chk_bytes()
+        chk_inv = CHKInventory.from_inventory(chk_bytes, inv)
+        self.assertFalse(chk_inv.has_id('fileid'))

=== modified file 'bzrlib/tests/test_pack_repository.py'
--- bzrlib/tests/test_pack_repository.py	2008-09-26 02:00:48 +0000
+++ bzrlib/tests/test_pack_repository.py	2008-10-09 04:50:37 +0000
@@ -511,12 +511,18 @@
             # can only stack on repositories that have compatible internal
             # metadata
             if getattr(repo._format, 'supports_tree_reference', False):
-                matching_format_name = 'pack-0.92-subtree'
+                if repo._format.supports_chks:
+                    matching_format_name = 'development3-subtree'
+                else:
+                    matching_format_name = 'pack-0.92-subtree'
             else:
                 matching_format_name = 'rich-root-pack'
             mismatching_format_name = 'pack-0.92'
         else:
-            matching_format_name = 'pack-0.92'
+            if repo._format.supports_chks:
+                matching_format_name = 'development3'
+            else:
+                matching_format_name = 'pack-0.92'
             mismatching_format_name = 'pack-0.92-subtree'
         base = self.make_repository('base', format=matching_format_name)
         repo.add_fallback_repository(base)
@@ -527,7 +533,7 @@
             repo.add_fallback_repository, bad_repo)
         self.assertContainsRe(str(e),
             r'(?m)KnitPackRepository.*/mismatch/.*\nis not compatible with\n'
-            r'KnitPackRepository.*/repo/.*\n'
+            r'.*Repository.*/repo/.*\n'
             r'different rich-root support')
 
     def test_stack_checks_serializers_compatibility(self):
@@ -535,7 +541,10 @@
         if getattr(repo._format, 'supports_tree_reference', False):
             # can only stack on repositories that have compatible internal
             # metadata
-            matching_format_name = 'pack-0.92-subtree'
+            if repo._format.supports_chks:
+                matching_format_name = 'development3-subtree'
+            else:
+                matching_format_name = 'pack-0.92-subtree'
             mismatching_format_name = 'rich-root-pack'
         else:
             if repo.supports_rich_root():
@@ -553,7 +562,7 @@
             repo.add_fallback_repository, bad_repo)
         self.assertContainsRe(str(e),
             r'(?m)KnitPackRepository.*/mismatch/.*\nis not compatible with\n'
-            r'KnitPackRepository.*/repo/.*\n'
+            r'.*Repository.*/repo/.*\n'
             r'different serializers')
 
     def test_adding_pack_does_not_record_pack_names_from_other_repositories(self):

=== modified file 'bzrlib/tests/test_repository.py'
--- bzrlib/tests/test_repository.py	2008-09-25 05:51:04 +0000
+++ bzrlib/tests/test_repository.py	2008-10-09 04:50:37 +0000
@@ -657,6 +657,29 @@
         self.assertFalse(repo._format.supports_external_lookups)
 
 
+class TestDevelopment3(TestCaseWithTransport):
+
+    def test_add_inventory_uses_chk_map(self):
+        repo = self.make_repository('repo', format="development3")
+        source = self.make_branch_and_tree("source", format="pack-0.92")
+        revid = source.commit("foo", rev_id="foo")
+        # get the inventory from the committed revision
+        basis = source.basis_tree()
+        basis.lock_read()
+        self.addCleanup(basis.unlock)
+        inv = basis.inventory
+        repo.lock_write()
+        self.addCleanup(repo.unlock)
+        repo.start_write_group()
+        self.addCleanup(repo.abort_write_group)
+        repo.add_inventory(revid, inv, [])
+        self.assertEqual(set([(revid,)]), repo.inventories.keys())
+        self.assertEqual(
+            set([('sha1:a521a815c343d72dffac50e316246f1fade1a4d3',),
+                ('sha1:a12ffb3c1810005e8ed9388d29602e1e9e8e06ac',)]),
+            repo.chk_bytes.keys())
+
+
 class TestWithBrokenRepo(TestCaseWithTransport):
     """These tests seem to be more appropriate as interface tests?"""
 

